from .events import (
    ClientEvent,
    ClientEventAuthentication,
    ClientEventChannelsListed,
    ClientEventHello,
    ClientEventIncompatibleVersion,
    ClientEventMessageReceived,
    ClientEventMessagesListed,
)
from .messages import (
    ClientMessageAuthenticate,
    ClientMessageHello,
    ClientMessageListChannels,
    ClientMessageListMessages,
    ClientMessagePost,
)
from .protocol import Client, ClientState



import concurrent.futures
import logging
import ssl
from pathlib import Path
from tkinter import BooleanVar, StringVar, messagebox
from tkinter.ttk import Button, Checkbutton, Entry, Frame, Label

from .app import TkApp
from .file_entry import ALL_FILETYPE, SSL_CERTIFICATE_FILETYPE, FileEntry

log = logging.getLogger(__name__)


class ConnectFrame(Frame):
    _connection_attempt: concurrent.futures.Future[None] | None

    def __init__(self, app: TkApp):
        super().__init__(padding=10)

        self.app = app

        self.grid_columnconfigure(1, weight=1)

        self.host_label = Label(self, text="Host:")
        self.host_label.grid(row=0, column=0, sticky="w")
        self.host_entry_var = StringVar(self)
        self.host_entry = Entry(self, textvariable=self.host_entry_var)
        self.host_entry.grid(row=0, column=1, sticky="ew")

        self.port_label = Label(self, text="Port:")
        self.port_label.grid(row=1, column=0, sticky="w")
        self.port_entry_var = StringVar(self)
        self.port_entry = Entry(self, textvariable=self.port_entry_var)
        self.port_entry.grid(row=1, column=1, sticky="ew")

        self.ssl_enabled_var = BooleanVar(self, value=False)
        self.ssl_enabled = Checkbutton(
            self,
            text="Use SSL?",
            variable=self.ssl_enabled_var,
        )
        self.ssl_enabled.grid(row=2, column=0, sticky="w")
        self.ssl_enabled_var.trace_add("write", self._on_cert_check_var_write)

        self.ssl_cert = FileEntry(
            self,
            text="Certificate (Optional):",
            browse_kwargs={
                "title": "Use Certificate File",
                "filetypes": [SSL_CERTIFICATE_FILETYPE, ALL_FILETYPE],
            },
        )
        self.ssl_cert.grid(row=2, column=1, padx=(15, 0), sticky="ew")

        self.nick_label = Label(self, text="Nickname:")
        self.nick_label.grid(row=3, column=0, sticky="w")
        self.nick_entry_var = StringVar(self)
        self.nick_entry = Entry(self, textvariable=self.nick_entry_var)
        self.nick_entry.grid(row=3, column=1, sticky="ew")

        self.connect = Button(self, text="Connect", command=self.do_connect)
        self.connect.grid(row=3, column=1, sticky="e")

        self.apply_last_connect_entries()

        self._connect_bind = app.bind("<Return>", lambda event: self.do_connect())

        self._connection_attempt = None
        self._destroying = False

    def apply_last_connect_entries(self) -> None:
        with self.app.store_factory() as store:
            host = store.get_setting("last-connect-host", "127.0.0.1")
            port = store.get_setting("last-connect-port", 6667)
            nick = store.get_setting("last-connect-nick", "Somebody")
            ssl_enabled = store.get_setting("last-connect-ssl-enabled", False)
            ssl_cert = store.get_setting("last-connect-ssl-cert", "")

        self.host_entry_var.set(host)
        self.port_entry_var.set(port)
        self.nick_entry_var.set(nick)
        self.ssl_enabled_var.set(ssl_enabled)
        self.ssl_cert.var.set(ssl_cert)

    def do_connect(self) -> None:
        fut = self._connection_attempt
        if fut is not None and not fut.done():
            return

        host = self.host_entry_var.get()
        port = int(self.port_entry_var.get())
        nick = self.nick_entry_var.get()
        ssl_enabled = self.ssl_enabled_var.get()
        ssl_cert = self.ssl_cert.get()

        if ssl_enabled:
            path = self.ssl_cert.get_path()
            if path is not None and not path.is_file():
                messagebox.showerror(
                    "Certificate Not Found",
                    "The SSL certificate path does not exist.",
                )
                return

            try:
                ssl_context = self._create_ssl_context(path)
            except ssl.SSLError as e:
                log.error("Could not create SSL context", exc_info=e)
                messagebox.showerror(
                    "Invalid Certificate",
                    "The SSL certificate could not be loaded.",
                )
                return

        else:
            ssl_context = None

        self.set_last_connect_entries(host, port, nick, ssl_enabled, ssl_cert)

        coro = self.app.attempt_connection(host, port, nick, ssl=ssl_context)
        self._connection_attempt = self.app.submit(coro)

        self.connect.state(["disabled"])
        self._connection_attempt.add_done_callback(self._on_connection_attempt)

    def set_last_connect_entries(
        self,
        host: str,
        port: int,
        nick: str,
        ssl_enabled: bool,
        ssl_cert: str,
    ) -> None:
        with self.app.store_factory() as store:
            store.set_setting("last-connect-host", host)
            store.set_setting("last-connect-port", port)
            store.set_setting("last-connect-nick", nick)
            store.set_setting("last-connect-ssl-enabled", ssl_enabled)
            store.set_setting("last-connect-ssl-cert", ssl_cert)

    def destroy(self) -> None:
        self._destroying = True
        self.app.unbind("<Return>", self._connect_bind)
        super().destroy()

    def _on_cert_check_var_write(self, *args) -> None:
        enabled = self.ssl_enabled_var.get()
        self.ssl_cert.enable(enabled)

    def _create_ssl_context(self, cafile: str | Path | None) -> ssl.SSLContext:
        if cafile is not None:
            cafile = str(cafile)

        context = ssl.create_default_context(cafile=cafile)

        # WARNING: for ease of server setup, skip hostname checks for self-signed certs
        if cafile is not None:
            context.check_hostname = False

        return context

    def _on_connection_attempt(self, fut: concurrent.futures.Future) -> None:
        if self._destroying:
            return

        self.connect.state(["!disabled"])



import contextlib
from typing import Iterator

from dumdum.protocol import varchar


class Reader:
    def __init__(self, buffer: bytearray | bytes) -> None:
        self.buffer = buffer
        self._index = 0
        self._closed = False

    def read(self, n: int = -1) -> bytes:
        if self._closed:
            raise RuntimeError("Cannot read from closed reader")

        if n < 0:
            n = len(self.buffer)
        else:
            n = min(n, len(self.buffer))

        start, self._index = self._index, self._index + n
        return self.buffer[start : self._index]

    def readexactly(self, n: int) -> bytes:
        if n < 0:
            raise ValueError(f"n must be 0 or greater, not {n}")

        data = self.read(n)
        if len(data) != n:
            raise IndexError(
                f"Insufficent data to read (expected {n}, got {len(data)})"
            )

        return data

    def read_bigint(self) -> int:
        data = self.readexactly(8)
        return int.from_bytes(data, byteorder="big")

    def read_varchar(self, *, max_length: int) -> str:
        return varchar.load(self, max_length=max_length)

    def close(self) -> None:
        self._closed = True


@contextlib.contextmanager
def bytearray_reader(buffer: bytearray) -> Iterator[Reader]:
    reader = Reader(buffer)

    try:
        yield reader
    finally:
        reader.close()

    buffer[: reader._index] = b""


@contextlib.contextmanager
def byte_reader(buffer: bytes) -> Iterator[Reader]:
    reader = Reader(buffer)

    try:
        yield reader
    finally:
        reader.close()



import contextlib
from typing import Any, Iterator, Self, TypeVar

from dumdum.appdirs import APP_DIRS
from dumdum.db import Connection, SQLiteConnection, run_migrations

T = TypeVar("T")


class ClientStore:
    def __init__(self, conn: Connection) -> None:
        self._conn = conn

    @contextlib.contextmanager
    def transaction(self) -> Iterator[Self]:
        with self._conn.transaction():
            yield self

    def get_last_selected_channel(
        self,
        addr: str,
        default: str | T = None,
    ) -> str | T:
        ret = self._conn.fetchval(
            "SELECT channel_name FROM last_selected_channel WHERE addr = ?",
            addr,
        )
        if ret is None:
            return default
        return ret

    def set_last_selected_channel(
        self, addr: str, channel_name: str | None = None
    ) -> None:
        self._conn.execute(
            "INSERT INTO last_selected_channel (addr, channel_name) VALUES (?1, ?2) "
            "ON CONFLICT (addr) DO UPDATE SET channel_name = ?2",
            addr,
            channel_name,
        )

    def get_setting(self, name: str, default: Any = None) -> Any:
        row = self._conn.fetchone("SELECT value FROM setting WHERE name = ?", name)
        if row is None:
            return default
        return row[0]

    def set_setting(self, name: str, value: Any) -> None:
        self._conn.execute(
            "INSERT INTO setting (name, value) VALUES (?1, ?2) "
            "ON CONFLICT (name) DO UPDATE SET value = ?2",
            name,
            value,
        )

    @classmethod
    @contextlib.contextmanager
    def from_appdirs(cls) -> Iterator[Self]:
        path = APP_DIRS.user_data_path / "client.db"
        with SQLiteConnection.connect(str(path)) as conn:
            with conn.transaction():
                run_migrations(conn, "client")

            yield cls(conn)



import math
from io import BytesIO
from typing import Protocol

from .errors import InvalidLengthError


class _Readable(Protocol):
    def read(self, n: int, /) -> bytes: ...


def load(f: _Readable, *, max_length: int) -> str:
    byte_count = math.ceil(max_length.bit_length() / 8)
    length_bytes = f.read(byte_count)
    if len(length_bytes) != byte_count:
        raise IndexError(f"Insufficient bytes for {max_length = }")

    length = int.from_bytes(length_bytes, byteorder="big")
    if length > max_length:
        raise InvalidLengthError(length, max_length)

    message = f.read(length)
    if length != len(message):
        raise IndexError(f"Insufficient bytes for {max_length = }")

    return message.decode()


def loads(message: bytes, *, max_length: int) -> str:
    return load(BytesIO(message), max_length=max_length)


def dumps(message: str, *, max_length: int) -> bytes:
    message_bytes = message.encode()

    length = len(message_bytes)
    if length > max_length:
        raise InvalidLengthError(length, max_length)

    byte_count = math.ceil(max_length.bit_length() / 8)
    length_bytes = len(message_bytes).to_bytes(byte_count, byteorder="big")

    return length_bytes + message_bytes



from dataclasses import dataclass
from typing import Self

from . import varchar
from .constants import MAX_CHANNEL_NAME_LENGTH, MAX_MESSAGE_LENGTH, MAX_NICK_LENGTH
from .reader import Reader


@dataclass
class Message:
    id: int
    channel_name: str
    nick: str
    content: str

    def __bytes__(self) -> bytes:
        return bytes(
            [
                *self.id.to_bytes(8, byteorder="big"),
                *varchar.dumps(self.channel_name, max_length=MAX_CHANNEL_NAME_LENGTH),
                *varchar.dumps(self.nick, max_length=MAX_NICK_LENGTH),
                *varchar.dumps(self.content, max_length=MAX_MESSAGE_LENGTH),
            ]
        )

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        id = reader.read_bigint()
        channel_name = reader.read_varchar(max_length=MAX_CHANNEL_NAME_LENGTH)
        nick = reader.read_varchar(max_length=MAX_NICK_LENGTH)
        content = reader.read_varchar(max_length=MAX_MESSAGE_LENGTH)
        return cls(id=id, channel_name=channel_name, nick=nick, content=content)



from enum import Enum
from typing import Sequence

from dumdum.protocol.buffer import extend_limited_buffer
from dumdum.protocol.channel import Channel
from dumdum.protocol.constants import (
    MAX_CHANNEL_NAME_LENGTH,
    MAX_MESSAGE_LENGTH,
    MAX_NICK_LENGTH,
)
from dumdum.protocol.enums import ClientMessageType
from dumdum.protocol.errors import InvalidStateError, MalformedDataError
from dumdum.protocol.interfaces import Protocol
from dumdum.protocol.message import Message
from dumdum.protocol.reader import Reader, bytearray_reader

from .events import (
    ServerEvent,
    ServerEventAuthentication,
    ServerEventHello,
    ServerEventIncompatibleVersion,
    ServerEventListChannels,
    ServerEventListMessages,
    ServerEventMessageReceived,
)
from .messages import (
    ServerMessageAcknowledgeAuthentication,
    ServerMessageHello,
    ServerMessageListChannels,
    ServerMessageListMessages,
    ServerMessagePost,
    ServerMessageSendIncompatibleVersion,
)

ParsedData = tuple[list[ServerEvent], bytes]


class ServerState(Enum):
    AWAITING_HELLO = 0
    AWAITING_AUTHENTICATION = 1
    READY = 2


class Server(Protocol):
    """The server for a single client."""

    PROTOCOL_VERSION = 2

    def __init__(self, *, buffer_size: int | None = 2**20) -> None:
        self.buffer_size = buffer_size

        self._buffer = bytearray()
        self._state = ServerState.AWAITING_HELLO

    def receive_bytes(self, data: bytes) -> ParsedData:
        extend_limited_buffer(self._buffer, data, limit=self.buffer_size)
        return self._maybe_parse_buffer()

    def hello(self, *, using_ssl: bool) -> bytes:
        self._assert_state(ServerState.AWAITING_AUTHENTICATION)
        return bytes(ServerMessageHello(using_ssl))

    def authenticate(self, *, success: bool) -> bytes:
        self._assert_state(ServerState.AWAITING_AUTHENTICATION)

        if success:
            self._state = ServerState.READY

        return bytes(ServerMessageAcknowledgeAuthentication(success))

    def send_message(self, message: Message) -> bytes:
        self._assert_state(ServerState.READY)
        return bytes(ServerMessagePost(message))

    def list_channels(self, channels: Sequence[Channel]) -> bytes:
        return bytes(ServerMessageListChannels(channels))

    def list_messages(self, messages: Sequence[Message]) -> bytes:
        return bytes(ServerMessageListMessages(messages))

    def _assert_state(self, *states: ServerState) -> None:
        if self._state not in states:
            raise InvalidStateError(self._state, states)

    def _maybe_parse_buffer(self) -> ParsedData:
        full_events: list[ServerEvent] = []
        full_outgoing = bytearray()

        try:
            while True:
                with bytearray_reader(self._buffer) as reader:
                    events, outgoing = self._read_message(reader)

                full_events.extend(events)
                full_outgoing.extend(outgoing)
        except (IndexError, ValueError) as e:
            # FIXME: this is making stuff hard to debug...
            if isinstance(e, UnicodeDecodeError):
                raise MalformedDataError(str(e)) from e

        return full_events, bytes(full_outgoing)

    def _read_message(self, reader: Reader) -> ParsedData:
        n = reader.readexactly(1)[0]
        try:
            t = ClientMessageType(n)
        except ValueError:
            raise MalformedDataError(f"Unknown message type {n}") from None

        if t == ClientMessageType.HELLO:
            return self._parse_hello(reader)
        elif t == ClientMessageType.AUTHENTICATE:
            return self._authenticate(reader)
        elif t == ClientMessageType.SEND_MESSAGE:
            return self._send_message(reader)
        elif t == ClientMessageType.LIST_CHANNELS:
            return self._list_channels(reader)
        elif t == ClientMessageType.LIST_MESSAGES:
            return self._list_messages(reader)

        raise RuntimeError(f"No handler for {t}")  # pragma: no cover

    def _parse_hello(self, reader: Reader) -> ParsedData:
        self._assert_state(ServerState.AWAITING_HELLO)

        version = reader.readexactly(1)[0]
        if version != self.PROTOCOL_VERSION:
            event = ServerEventIncompatibleVersion(version)
            response = ServerMessageSendIncompatibleVersion(self.PROTOCOL_VERSION)
            return [event], bytes(response)

        event = ServerEventHello()
        self._state = ServerState.AWAITING_AUTHENTICATION
        return [event], b""

    def _authenticate(self, reader: Reader) -> ParsedData:
        self._assert_state(ServerState.AWAITING_AUTHENTICATION)
        nick = reader.read_varchar(max_length=MAX_NICK_LENGTH)
        event = ServerEventAuthentication(nick=nick)
        return [event], b""

    def _send_message(self, reader: Reader) -> ParsedData:
        self._assert_state(ServerState.READY)
        channel_name = reader.read_varchar(max_length=MAX_CHANNEL_NAME_LENGTH)
        content = reader.read_varchar(max_length=MAX_MESSAGE_LENGTH)

        event = ServerEventMessageReceived(channel_name, content)
        # TODO: broadcast message to all users
        return [event], b""

    def _list_channels(self, reader: Reader) -> ParsedData:
        self._assert_state(ServerState.READY)
        event = ServerEventListChannels()
        return [event], b""

    def _list_messages(self, reader: Reader) -> ParsedData:
        self._assert_state(ServerState.READY)
        channel_name = reader.read_varchar(max_length=MAX_CHANNEL_NAME_LENGTH)
        before = reader.read_bigint() or None
        after = reader.read_bigint() or None
        event = ServerEventListMessages(channel_name, before, after)
        return [event], b""



from enum import Enum


class ClientMessageType(Enum):
    HELLO = 0
    AUTHENTICATE = 2
    SEND_MESSAGE = 3
    LIST_CHANNELS = 4
    LIST_MESSAGES = 5


class ServerMessageType(Enum):
    HELLO = 0
    INCOMPATIBLE_VERSION = 1
    ACKNOWLEDGE_AUTHENTICATION = 2
    SEND_MESSAGE = 3
    LIST_CHANNELS = 4
    LIST_MESSAGES = 5



from .events import (
    ServerEvent,
    ServerEventAuthentication,
    ServerEventHello,
    ServerEventIncompatibleVersion,
    ServerEventListChannels,
    ServerEventListMessages,
    ServerEventMessageReceived,
)
from .messages import (
    ServerMessageAcknowledgeAuthentication,
    ServerMessageHello,
    ServerMessageListChannels,
    ServerMessageListMessages,
    ServerMessagePost,
    ServerMessageSendIncompatibleVersion,
)
from .protocol import Server, ServerState



import pytest

from dumdum.protocol import bytearray_reader, byte_reader


def test_bytearray_reader_commit_and_rollback():
    data = bytearray(b"Hello world!\n")

    with bytearray_reader(data) as reader:
        assert reader.read(5) == b"Hello"

    assert data == b" world!\n"

    with pytest.raises(RuntimeError), bytearray_reader(data) as reader:
        assert reader.read() == b" world!\n"
        raise RuntimeError

    assert data == b" world!\n"


def test_reader_out_of_bounds():
    with byte_reader(b"Hello world!\n") as reader:
        with pytest.raises(ValueError):
            reader.readexactly(-1)

        assert reader.read(1000) == b"Hello world!\n"

        reader.readexactly(0)

        with pytest.raises(IndexError):
            reader.readexactly(1)


def test_reader_varchar():
    with byte_reader(b"\x05Hello\x08 world!\n") as reader:
        assert reader.read_varchar(max_length=5) == "Hello"
        assert reader.read_varchar(max_length=8) == " world!\n"


def test_reader_closed():
    with byte_reader(b"Hello world!\n") as reader:
        pass

    with pytest.raises(RuntimeError):
        reader.read()

    with pytest.raises(RuntimeError):
        reader.readexactly(1)



class ClientCannotUpgradeSSLError(Exception):
    """Raised when the server requires SSL and the client cannot upgrade."""


class ServerCannotUpgradeSSLError(Exception):
    """Raised when the server does not use SSL and the client requires it."""


class AuthenticationFailedError(Exception):
    """Raised when authentication with the dumdum server fails."""



from dataclasses import dataclass
from typing import Sequence

from dumdum.protocol.channel import Channel
from dumdum.protocol.constants import (
    MAX_LIST_CHANNEL_LENGTH_BYTES,
    MAX_LIST_MESSAGE_LENGTH_BYTES,
)
from dumdum.protocol.enums import ServerMessageType
from dumdum.protocol.message import Message


@dataclass
class ServerMessageHello:
    using_ssl: bool

    def __bytes__(self) -> bytes:
        return bytes([ServerMessageType.HELLO.value, self.using_ssl])


@dataclass
class ServerMessageSendIncompatibleVersion:
    required: int

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ServerMessageType.INCOMPATIBLE_VERSION.value,
                self.required,
            ]
        )


@dataclass
class ServerMessageAcknowledgeAuthentication:
    success: bool

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ServerMessageType.ACKNOWLEDGE_AUTHENTICATION.value,
                self.success,
            ]
        )


@dataclass
class ServerMessagePost:
    message: Message

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ServerMessageType.SEND_MESSAGE.value,
                *bytes(self.message),
            ]
        )


@dataclass
class ServerMessageListChannels:
    channels: Sequence[Channel]

    def __bytes__(self) -> bytes:
        channel_bytes = b"".join(bytes(c) for c in self.channels)
        channel_length = len(channel_bytes).to_bytes(
            MAX_LIST_CHANNEL_LENGTH_BYTES,
            byteorder="big",
        )
        return bytes(
            [
                ServerMessageType.LIST_CHANNELS.value,
                *channel_length,
                *channel_bytes,
            ]
        )


@dataclass
class ServerMessageListMessages:
    messages: Sequence[Message]

    def __bytes__(self) -> bytes:
        message_bytes = b"".join(bytes(c) for c in self.messages)
        message_length = len(message_bytes).to_bytes(
            MAX_LIST_MESSAGE_LENGTH_BYTES,
            byteorder="big",
        )
        return bytes(
            [
                ServerMessageType.LIST_MESSAGES.value,
                *message_length,
                *message_bytes,
            ]
        )



from abc import ABC, abstractmethod
from typing import Sequence


class Protocol(ABC):
    @abstractmethod
    def receive_bytes(self, data: bytes) -> tuple[Sequence[object], bytes]:
        """Receive bytes from the sender."""



from enum import Enum

from dumdum.protocol.buffer import extend_limited_buffer
from dumdum.protocol.channel import Channel
from dumdum.protocol.constants import (
    MAX_LIST_CHANNEL_LENGTH_BYTES,
    MAX_LIST_MESSAGE_LENGTH_BYTES,
)
from dumdum.protocol.enums import ServerMessageType
from dumdum.protocol.errors import InvalidStateError, MalformedDataError
from dumdum.protocol.interfaces import Protocol
from dumdum.protocol.message import Message
from dumdum.protocol.reader import Reader, byte_reader, bytearray_reader

from .events import (
    ClientEvent,
    ClientEventAuthentication,
    ClientEventChannelsListed,
    ClientEventHello,
    ClientEventIncompatibleVersion,
    ClientEventMessageReceived,
    ClientEventMessagesListed,
)
from .messages import (
    ClientMessageAuthenticate,
    ClientMessageHello,
    ClientMessageListChannels,
    ClientMessageListMessages,
    ClientMessagePost,
)

ParsedData = tuple[list[ClientEvent], bytes]


class ClientState(Enum):
    AWAITING_HELLO = 0
    AWAITING_AUTHENTICATION = 1
    READY = 2


class Client(Protocol):
    """The client connected to a server."""

    PROTOCOL_VERSION = 2

    def __init__(self, nick: str, *, buffer_size: int | None = 2**20) -> None:
        self.nick = nick
        self.buffer_size = buffer_size

        self._buffer = bytearray()
        self._state = ClientState.AWAITING_HELLO

    def receive_bytes(self, data: bytes) -> ParsedData:
        extend_limited_buffer(self._buffer, data, limit=self.buffer_size)
        return self._maybe_parse_buffer()

    def hello(self) -> bytes:
        self._assert_state(ClientState.AWAITING_HELLO)
        return bytes(ClientMessageHello(self.PROTOCOL_VERSION))

    def authenticate(self) -> bytes:
        self._assert_state(ClientState.AWAITING_AUTHENTICATION)
        return bytes(ClientMessageAuthenticate(self.nick))

    def send_message(self, channel_name: str, content: str) -> bytes:
        self._assert_state(ClientState.READY)
        return bytes(ClientMessagePost(channel_name, content))

    def list_channels(self) -> bytes:
        self._assert_state(ClientState.READY)
        return bytes(ClientMessageListChannels())

    def list_messages(
        self,
        channel_name: str,
        *,
        before: int | None = None,
        after: int | None = None,
    ) -> bytes:
        self._assert_state(ClientState.READY)

        if before is not None and before < 1:
            raise ValueError(f"before must be 1 or greater, not {before}")
        if after is not None and after < 1:
            raise ValueError(f"after must be 1 or greater, not {after}")

        return bytes(ClientMessageListMessages(channel_name, before, after))

    def _assert_state(self, *states: ClientState) -> None:
        if self._state not in states:
            raise InvalidStateError(self._state, states)

    def _maybe_parse_buffer(self) -> ParsedData:
        full_events: list[ClientEvent] = []
        full_outgoing = bytearray()

        try:
            while True:
                with bytearray_reader(self._buffer) as reader:
                    events, outgoing = self._read_message(reader)

                full_events.extend(events)
                full_outgoing.extend(outgoing)
        except (IndexError, ValueError) as e:
            # FIXME: this is making stuff hard to debug...
            if isinstance(e, UnicodeDecodeError):
                raise MalformedDataError(str(e)) from e

        return full_events, bytes(full_outgoing)

    def _read_message(self, reader: Reader) -> ParsedData:
        n = reader.readexactly(1)[0]
        try:
            t = ServerMessageType(n)
        except ValueError:
            raise MalformedDataError(f"Unknown message type {n}") from None

        if t == ServerMessageType.HELLO:
            return self._parse_hello(reader)
        elif t == ServerMessageType.INCOMPATIBLE_VERSION:
            return self._parse_incompatible_version(reader)
        elif t == ServerMessageType.ACKNOWLEDGE_AUTHENTICATION:
            return self._accept_authentication(reader)
        elif t == ServerMessageType.SEND_MESSAGE:
            return self._parse_message(reader)
        elif t == ServerMessageType.LIST_CHANNELS:
            return self._parse_channel_list(reader)
        elif t == ServerMessageType.LIST_MESSAGES:
            return self._parse_message_list(reader)

        raise RuntimeError(f"No handler for {t}")  # pragma: no cover

    def _parse_hello(self, reader: Reader) -> ParsedData:
        self._assert_state(ClientState.AWAITING_HELLO)

        using_ssl = int.from_bytes(reader.readexactly(1), byteorder="big") > 0
        event = ClientEventHello(using_ssl)

        self._state = ClientState.AWAITING_AUTHENTICATION
        return [event], b""

    def _parse_incompatible_version(self, reader: Reader) -> ParsedData:
        self._assert_state(ClientState.AWAITING_HELLO)
        version = reader.readexactly(1)[0]
        event = ClientEventIncompatibleVersion(version, self.PROTOCOL_VERSION)
        return [event], b""

    def _accept_authentication(self, reader: Reader) -> ParsedData:
        self._assert_state(ClientState.AWAITING_AUTHENTICATION)

        success = reader.readexactly(1)[0] > 0
        if success:
            self._state = ClientState.READY

        event = ClientEventAuthentication(success)
        return [event], b""

    def _parse_message(self, reader: Reader) -> ParsedData:
        self._assert_state(ClientState.READY)
        message = Message.from_reader(reader)
        event = ClientEventMessageReceived(message)
        return [event], b""

    def _parse_channel_list(self, reader: Reader) -> ParsedData:
        self._assert_state(ClientState.READY)
        length = int.from_bytes(
            reader.readexactly(MAX_LIST_CHANNEL_LENGTH_BYTES),
            byteorder="big",
        )
        channel_bytes = reader.readexactly(length)

        channels: list[Channel] = []
        with byte_reader(channel_bytes) as channel_reader:
            try:
                while True:
                    channels.append(Channel.from_reader(channel_reader))
            except IndexError:
                pass

        event = ClientEventChannelsListed(channels)
        return [event], b""

    def _parse_message_list(self, reader: Reader) -> ParsedData:
        self._assert_state(ClientState.READY)
        length = int.from_bytes(
            reader.readexactly(MAX_LIST_MESSAGE_LENGTH_BYTES),
            byteorder="big",
        )
        message_bytes = reader.readexactly(length)

        messages: list[Message] = []
        with byte_reader(message_bytes) as message_reader:
            try:
                while True:
                    messages.append(Message.from_reader(message_reader))
            except IndexError:
                pass

        event = ClientEventMessagesListed(messages)
        return [event], b""



from .errors import BufferOverflowError


def extend_limited_buffer(
    buffer: bytearray,
    data: bytes | bytearray,
    *,
    limit: int | None,
) -> None:
    if limit is None:
        buffer.extend(data)
        return

    len_buffer, len_data = len(buffer), len(data)
    if len_buffer + len_data > limit:
        raise BufferOverflowError(limit, len_buffer, len_data)

    buffer.extend(data)



import asyncio
import contextlib
import logging
import ssl

from dumdum.protocol import (
    InvalidStateError,
    Message,
    Server,
    ServerEvent,
    ServerEventAuthentication,
    ServerEventHello,
    ServerEventListChannels,
    ServerEventListMessages,
    ServerEventMessageReceived,
    create_snowflake,
)

from .connection import Connection
from .state import ServerState

log = logging.getLogger(__name__)


class Manager:
    def __init__(
        self,
        state: ServerState,
        ssl: ssl.SSLContext | None,
        *,
        drain_timeout: float = 30,
        close_timeout: float = 5,
    ) -> None:
        self.state = state
        self.connections: list[Connection] = []
        self.ssl = ssl
        self.drain_timeout = drain_timeout
        self.close_timeout = close_timeout

    async def accept_connection(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ):
        addr = writer.get_extra_info("peername")
        log.info("Accepted connection from %s", addr)

        connection = Connection(self, reader, writer, self._create_server())
        self.connections.append(connection)
        try:
            await connection.communicate()
        except asyncio.CancelledError:
            # Don't need to log this exception
            asyncio.current_task().uncancel()  # type: ignore
        except ConnectionResetError:
            # Client wants to disconnect
            pass
        except BaseException:
            log.exception("Error occurred while handling %s", addr)
        finally:
            log.info("Connection %s has disconnected", addr)

            writer.close()
            await self._wait_closed(writer)

            self._close_connection(connection)

    def _create_server(self) -> Server:
        return Server()

    async def _wait_closed(self, writer: asyncio.StreamWriter) -> None:
        timeout = self.close_timeout
        with contextlib.suppress(Exception):
            await asyncio.wait_for(writer.wait_closed(), timeout=timeout)

    async def _handle_events(self, conn: Connection, events: list[ServerEvent]) -> None:
        for event in events:
            await self._handle_event(conn, event)

    async def _handle_event(self, conn: Connection, event: ServerEvent) -> None:
        log.debug(
            "%s produced by %s",
            type(event).__name__,
            conn.writer.get_extra_info("peername"),
        )

        if isinstance(event, ServerEventHello):
            await self._hello(conn, event)
        elif isinstance(event, ServerEventAuthentication):
            self._authenticate(conn, event)
        elif isinstance(event, ServerEventMessageReceived):
            self._broadcast_message(conn, event)
        elif isinstance(event, ServerEventListChannels):
            self._list_channels(conn, event)
        elif isinstance(event, ServerEventListMessages):
            self._list_messages(conn, event)

    async def _hello(self, conn: Connection, event: ServerEventHello) -> None:
        using_ssl = self.ssl is not None

        data = conn.server.hello(using_ssl=using_ssl)
        conn.writer.write(data)

        if using_ssl:
            assert self.ssl is not None
            await conn.writer.start_tls(self.ssl)

    def _authenticate(self, conn: Connection, event: ServerEventAuthentication) -> None:
        user = self.state.get_user(event.nick)
        if user is None:
            self.state.add_user(event.nick)
            conn.nick = event.nick
            success = True
        else:
            success = False

        data = conn.server.authenticate(success=success)
        conn.writer.write(data)

    def _broadcast_message(
        self,
        conn: Connection,
        event: ServerEventMessageReceived,
    ) -> None:
        assert conn.nick is not None

        if self.state.get_channel(event.channel_name) is None:
            return

        message = Message(
            create_snowflake(),
            event.channel_name,
            conn.nick,
            event.content,
        )
        self.state.add_message(message)

        for peer in self.connections:
            with contextlib.suppress(InvalidStateError):
                data = peer.server.send_message(message)
                peer.writer.write(data)

    def _list_channels(self, conn: Connection, event: ServerEventListChannels) -> None:
        data = conn.server.list_channels(self.state.channels)
        conn.writer.write(data)

    def _list_messages(self, conn: Connection, event: ServerEventListMessages) -> None:
        messages = self.state.get_messages(
            event.channel_name,
            before=event.before,
            after=event.after,
        )
        data = conn.server.list_messages(messages)
        conn.writer.write(data)

    def _close_connection(self, conn: Connection) -> None:
        self.connections.remove(conn)
        if conn.nick is not None:
            self.state.remove_user(conn.nick)


async def host_server(
    state: ServerState,
    host: str | None,
    port: int,
    *,
    ssl: ssl.SSLContext | None,
) -> None:
    manager = Manager(state, ssl)
    server = await asyncio.start_server(
        manager.accept_connection,
        host=host,
        port=port,
    )
    async with server:
        await server.serve_forever()



import contextlib
import sqlite3
from abc import ABC, abstractmethod
from typing import Any, ContextManager, Iterator, Protocol, Self, Sequence

Row = Sequence[Any]


class Cursor(Protocol):
    def fetchone(self) -> Row: ...
    def fetchall(self) -> Sequence[Row]: ...


class Connection(ABC):
    @abstractmethod
    def execute(self, query: str, /, *parameters: Any) -> Cursor:
        """Execute query and return a cursor."""

    @abstractmethod
    def executescript(self, query: str, /) -> None:
        """Execute one or more SQL statements."""

    @abstractmethod
    def transaction(self) -> ContextManager[Self]:
        """Returns a context manager to begin a transaction."""

    def fetchone(self, query: str, /, *parameters: Any) -> Row | None:
        """Execute query and return the first row.

        If the results are empty, None is returned.

        """
        c = self.execute(query, *parameters)
        return c.fetchone()

    def fetchall(self, query: str, /, *parameters: Any) -> Sequence[Row]:
        """Execute query and return all rows.

        If the results are empty, None is returned.

        """
        c = self.execute(query, *parameters)
        return c.fetchall()

    def fetchval(self, query: str, /, *parameters: Any) -> Any:
        """Execute query and return the first value of the first row.

        If the results are empty, None is returned.

        """
        row = self.fetchone(query, *parameters)
        if row is not None:
            return row[0]


class SQLiteConnection(Connection):
    def __init__(self, conn: sqlite3.Connection) -> None:
        self._conn = conn

    def execute(self, query: str, *parameters: Any) -> sqlite3.Cursor:
        if len(parameters) == 1 and isinstance(parameters[0], (tuple, list, dict)):
            _params = parameters[0]
        else:
            _params = parameters

        return self._conn.execute(query, _params)

    def executescript(self, query: str) -> None:
        self._conn.executescript(query)

    @contextlib.contextmanager
    def transaction(self) -> Iterator[Self]:
        with self._conn:
            yield self

    @classmethod
    @contextlib.contextmanager
    def connect(cls, *args, **kwargs) -> Iterator[Self]:
        with contextlib.closing(sqlite3.connect(*args, **kwargs)) as conn:
            yield cls(conn)



import asyncio
import concurrent.futures
import logging
import queue
import ssl
from tkinter import Event, Tk, messagebox
from tkinter.ttk import Frame
from typing import Any, Awaitable, ContextManager, Protocol, Type, runtime_checkable

from dumdum.protocol import (
    ClientEvent,
    ClientEventAuthentication,
    ClientEventIncompatibleVersion,
)

from .async_client import AsyncClient
from .errors import (
    AuthenticationFailedError,
    ClientCannotUpgradeSSLError,
    ServerCannotUpgradeSSLError,
)
from .event_thread import EventThread
from .store import ClientStore

log = logging.getLogger(__name__)


def has_exception(
    exc: BaseException,
    base: Type[BaseException] | tuple[Type[BaseException], ...],
) -> bool:
    if isinstance(exc, BaseExceptionGroup):
        return exc.subgroup(base) is not None
    return isinstance(exc, base)


def is_self_signed_certificate_exception(exc: BaseException) -> bool:
    def predicate(exc: BaseException) -> bool:
        if not isinstance(exc, ssl.SSLCertVerificationError):
            return False
        # https://www.openssl.org/docs/man1.0.2/man1/verify.html
        return exc.verify_code == 18

    if not isinstance(exc, BaseExceptionGroup):
        exc = BaseExceptionGroup("", [exc])
    return exc.subgroup(predicate) is not None


class ClientStoreFactory(Protocol):
    def __call__(self) -> ContextManager[ClientStore]: ...


class TkApp(Tk):
    client: AsyncClient  # NOTE: only assigned in attempt_connection()
    _client_events: queue.Queue[ClientEvent]
    _last_connection_exc: BaseException | None

    def __init__(
        self,
        event_thread: EventThread,
        store_factory: ClientStoreFactory,
    ):
        super().__init__()

        self.event_thread = event_thread
        self.store_factory = store_factory

        self._connect_lifetime_with_event_thread(event_thread)

        self._client_events = queue.Queue()
        self._last_connection_exc = None

        self.title("Dumdum Client")
        self.geometry("900x600")

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.frame = Frame(self)
        self.frame.grid()

        self.bind("<<Destroy>>", self._on_destroy)
        self.bind("<<ClientEvent>>", self._on_client_event)
        self.bind("<<ConnectionLost>>", self._on_connection_lost)

    def switch_frame(self, frame: Frame) -> None:
        self.frame.destroy()
        self.frame = frame
        self.frame.grid(sticky="nesw")

    def submit(self, coro: Awaitable[Any]) -> concurrent.futures.Future:
        fut = asyncio.run_coroutine_threadsafe(coro, self.event_thread.loop)
        fut.add_done_callback(log_fut_exception)
        return fut

    async def attempt_connection(
        self,
        host: str,
        port: int,
        nick: str,
        *,
        ssl: ssl.SSLContext | None,
    ) -> None:
        self.client = AsyncClient(nick, event_callback=self._handle_event_threadsafe)

        coro = self._run_connection(host, port, ssl=ssl)
        self._connection_task = asyncio.create_task(coro)

        await self.client._wait_for_authentication()

    def _connect_lifetime_with_event_thread(self, event_thread: EventThread) -> None:
        # In our application we'll be running an asyncio event loop in
        # a separate thread. This event loop may try to run methods on
        # our GUI like event_generate(), which requires the GUI to be running.
        # If the GUI is destroyed first, it may cause a deadlock
        # in the other thread, preventing our program from exiting.
        # As such, we need to defer GUI destruction until the event thread
        # is finished.
        event_callback = lambda fut: self.event_generate("<<Destroy>>")
        event_thread.finished_fut.add_done_callback(event_callback)

    def destroy(self) -> None:
        self.event_thread.stop()

    def _on_destroy(self, event: Event) -> None:
        super().destroy()

    async def _run_connection(
        self,
        host: str,
        port: int,
        *,
        ssl: ssl.SSLContext | None,
    ) -> None:
        try:
            async with self.client.connect(host, port, ssl=ssl):
                await self.client.run_forever()
        except BaseException as e:
            self._last_connection_exc = e
        else:
            self._last_connection_exc = None
        finally:
            self.event_generate("<<ConnectionLost>>")

    def _handle_event_threadsafe(self, event: ClientEvent):
        self._client_events.put_nowait(event)
        self.event_generate("<<ClientEvent>>")

    def _on_client_event(self, event: Event) -> None:
        self._handle_event(self._client_events.get_nowait())

    def _handle_event(self, event: ClientEvent) -> None:
        if isinstance(event, ClientEventIncompatibleVersion):
            message = (
                f"The server's protocol version does not match our client version. "
                f"We require V{event.client_version}, but the server "
                f"requires V{event.server_version}."
            )
            messagebox.showerror("Incompatible Version", message)
        elif isinstance(event, ClientEventAuthentication):
            if not event.success:
                message = (
                    "There was an issue connecting to the server. "
                    "Maybe your nickname was taken?"
                )
                messagebox.showerror("Authentication Interrupted", message)
                return

            from .chat_frame import ChatFrame

            self.switch_frame(ChatFrame(self))
            self.submit(self.client.list_channels())

        if isinstance(self.frame, Dispachable):
            self.frame.handle_client_event(event)

    def _on_connection_lost(self, event: Event) -> None:
        exc = self._last_connection_exc
        if exc is None:
            log.info("Disconnected from server")
            messagebox.showwarning(
                "Connection Lost",
                "We have been disconnected by the server.",
            )
        elif isinstance(exc, asyncio.CancelledError):
            # The user wants to close the GUI
            return
        elif has_exception(exc, AuthenticationFailedError):
            # Handled via ClientEventAuthentication
            return
        elif has_exception(exc, ClientCannotUpgradeSSLError):
            log.info("Cannot connect to server, enabling SSL is required")
            messagebox.showerror(
                "Cannot Upgrade SSL",
                "The server requires SSL encryption. You must turn on SSL, "
                "which may require a self-signed certificate from them.",
            )
        elif has_exception(exc, ServerCannotUpgradeSSLError):
            log.info("Cannot connect to server, disabling SSL is required")
            messagebox.showerror(
                "Cannot Upgrade SSL",
                "The server is unable to use SSL encryption. If you still want to "
                "connect using an insecure connection, you must turn off SSL.",
            )
        elif is_self_signed_certificate_exception(exc):
            log.info("Cannot connect to server, self-signed certificate must be provided")
            messagebox.showerror(
                "Cannot Upgrade SSL",
                "The server is using a self-signed certificate. In order to "
                "connect, you must download their certificate from a trusted "
                "source and then specify it in the certificate field.",
            )
        elif isinstance(exc, BaseExceptionGroup) and len(exc.exceptions) == 1:
            first_exception = exc.exceptions[0]
            log.error("Lost connection with server", exc_info=exc)
            messagebox.showerror("Connection Lost", str(first_exception))
        elif isinstance(exc, BaseExceptionGroup):
            log.error("Lost connection with server", exc_info=exc)
            messagebox.showerror(
                "Connection Lost",
                "Multiple errors occurred during the connection. "
                "See logs for more information.",
            )
        else:
            log.error("Lost connection with server", exc_info=exc)
            messagebox.showerror("Connection Lost", str(exc))

        from .connect_frame import ConnectFrame

        if not isinstance(self.frame, ConnectFrame):
            self.switch_frame(ConnectFrame(self))


def log_fut_exception(fut: concurrent.futures.Future) -> None:
    if fut.cancelled():
        return

    exc = fut.exception()
    if exc is not None:
        log.error("Uncaught exception from future %r", fut, exc_info=exc)


@runtime_checkable
class Dispachable(Protocol):
    def handle_client_event(self, event: ClientEvent) -> Any: ...



from enum import Enum


class ProtocolError(Exception):
    """The base class for errors related to the protocol."""


class BufferOverflowError(ProtocolError):
    """The protocol received more data than the buffer could store.

    This may be the result of the buffer size being set too low,
    or possibly malformed data that the protocol is unable to parse.

    Clients and servers, upon receiving this error, should immediately
    terminate their connection.

    """

    def __init__(self, limit: int, len_buffer: int, len_data: int) -> None:
        super().__init__(f"Buffer limit cannot be exceeded ({limit:,d} bytes)")
        self.limit = limit
        self.len_buffer = len_buffer
        self.len_data = len_data


class MalformedDataError(ProtocolError):
    """Raised when there is unambiguously malformed data.

    Clients and servers, upon receiving this error, should immediately
    terminate their connection.

    """


class InvalidLengthError(MalformedDataError):
    """An invalid length was specified for a variable-length message."""

    def __init__(self, length: int, max_length: int) -> None:
        super().__init__(
            f"Expected message length up to {max_length}, got {length} instead"
        )
        self.length = length
        self.max_length = max_length


class InvalidStateError(ProtocolError):
    """Some action cannot be handled in the protocol's current state."""

    def __init__(self, current_state: Enum, expected_states: tuple[Enum, ...]) -> None:
        super().__init__(
            "{} must be {}, but is currently {}".format(
                type(current_state).__name__,
                " or ".join(s.name for s in expected_states),
                current_state.name,
            )
        )
        self.current_state = current_state
        self.expected_states = expected_states



import asyncio
import contextlib
import ssl
from typing import Any, AsyncIterator, Callable, Iterator, Self

from dumdum.protocol import (
    Client,
    ClientEvent,
    ClientEventAuthentication,
    ClientEventHello,
    ClientEventIncompatibleVersion,
)

from .errors import (
    AuthenticationFailedError,
    ClientCannotUpgradeSSLError,
    ServerCannotUpgradeSSLError,
)


def maybe_create_fut(last: asyncio.Future | None) -> asyncio.Future:
    if last is None:
        return asyncio.get_running_loop().create_future()
    return last


class AsyncClient:
    _reader: asyncio.StreamReader | None
    _writer: asyncio.StreamWriter | None
    _read_task: asyncio.Task | None
    _addr: str | None
    _auth_fut: asyncio.Future[bool | None] | None
    _ssl_context: ssl.SSLContext | None

    def __init__(
        self,
        nick: str,
        *,
        event_callback: Callable[[ClientEvent], Any],
        drain_timeout: float = 30,
        close_timeout: float = 5,
    ) -> None:
        self.nick = nick
        self.event_callback = event_callback
        self.drain_timeout = drain_timeout
        self.close_timeout = close_timeout

        self._protocol = Client(nick)
        self._reader = None
        self._writer = None
        self._read_task = None
        self._addr = None

        self._auth_fut = None
        self._ssl_context = None

    @property
    def addr(self) -> str:
        if self._addr is None:
            raise RuntimeError("addr is not yet defined")
        return self._addr

    @contextlib.asynccontextmanager
    async def connect(
        self,
        host: str,
        port: int,
        *,
        ssl: ssl.SSLContext | None,
    ) -> AsyncIterator[Self]:
        self._addr = f"{host}:{port}"  # FIXME: must be canonicalized
        self._ssl_context = ssl

        with self._prepare_auth_fut():
            self._reader, self._writer = await asyncio.open_connection(host, port)
            async with asyncio.TaskGroup() as tg:
                _read_coro = self._read_loop(self._reader, self._writer)
                self._read_task = tg.create_task(_read_coro)

                try:
                    success = await self._handshake()
                    if not success:
                        raise AuthenticationFailedError()

                    yield self
                finally:
                    await self.close()

    async def run_forever(self) -> None:
        assert self._read_task is not None
        await self._read_task

    async def close(self) -> None:
        if self._writer is None:
            return

        # Any exceptions here will be repeated in _read_loop()
        self._writer.close()
        await self._wait_closed()

    async def send_message(self, channel_name: str, content: str) -> None:
        data = self._protocol.send_message(channel_name, content)
        await self._send_and_drain(data)

    async def list_channels(self) -> None:
        data = self._protocol.list_channels()
        await self._send_and_drain(data)

    async def list_messages(
        self,
        channel_name: str,
        *,
        before: int | None = None,
        after: int | None = None,
    ) -> None:
        data = self._protocol.list_messages(channel_name, before=before, after=after)
        await self._send_and_drain(data)

    @contextlib.contextmanager
    def _prepare_auth_fut(self) -> Iterator[None]:
        self._auth_fut = maybe_create_fut(self._auth_fut)
        try:
            yield
        finally:
            self._set_authentication(None)

    async def _read_loop(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        while True:
            data = await reader.read(1024)
            if len(data) == 0:
                break

            events, outgoing = self._protocol.receive_bytes(data)
            writer.write(outgoing)
            await self._handle_events(events)
            await self._drain()  # exert backpressure

    async def _handshake(self) -> bool | None:
        assert self._writer is not None
        data = self._protocol.hello()
        self._writer.write(data)
        return await self._wait_for_authentication()

    async def _wait_for_authentication(self) -> bool | None:
        self._auth_fut = maybe_create_fut(self._auth_fut)
        return await asyncio.shield(self._auth_fut)

    async def _handle_events(self, events: list[ClientEvent]) -> None:
        for event in events:
            await self._handle_event(event)

    async def _handle_event(self, event: ClientEvent) -> None:
        if isinstance(event, ClientEventHello):
            if event.using_ssl:
                await self._upgrade_to_ssl()
            elif self._ssl_context is not None:
                raise ServerCannotUpgradeSSLError()

            assert self._writer is not None
            data = self._protocol.authenticate()
            self._writer.write(data)
        elif isinstance(event, ClientEventIncompatibleVersion):
            assert self._writer is not None
            self._writer.close()
        elif isinstance(event, ClientEventAuthentication):
            self._set_authentication(event.success)
        self._dispatch_event(event)

    async def _upgrade_to_ssl(self) -> None:
        if self._ssl_context is None:
            raise ClientCannotUpgradeSSLError()

        assert self._writer is not None
        await self._writer.start_tls(self._ssl_context)

    def _set_authentication(self, result: bool | None) -> None:
        assert self._auth_fut is not None
        if not self._auth_fut.done():
            self._auth_fut.set_result(result)

    def _dispatch_event(self, event: ClientEvent) -> None:
        self.event_callback(event)

    async def _send_and_drain(self, data: bytes) -> None:
        assert self._writer is not None
        self._writer.write(data)
        await self._drain()

    async def _drain(self) -> None:
        assert self._writer is not None
        await asyncio.wait_for(self._writer.drain(), timeout=self.drain_timeout)

    async def _wait_closed(self) -> None:
        assert self._writer is not None
        timeout = self.close_timeout
        with contextlib.suppress(Exception):
            await asyncio.wait_for(self._writer.wait_closed(), timeout=timeout)



from dataclasses import dataclass


@dataclass
class ServerEvent:
    """An event received by the server."""


@dataclass
class ServerEventHello(ServerEvent):
    """The client says hello."""


@dataclass
class ServerEventIncompatibleVersion(ServerEvent):
    """The client tried using an incompatible protocol version."""

    version: int


@dataclass
class ServerEventAuthentication(ServerEvent):
    """The client attempted to authenticate with the server."""

    nick: str


@dataclass
class ServerEventMessageReceived(ServerEvent):
    """The client sent a message to the server."""

    channel_name: str
    content: str


@dataclass
class ServerEventListChannels(ServerEvent):
    """The client requested a list of channels."""


@dataclass
class ServerEventListMessages(ServerEvent):
    """The client requested a list of messages."""

    channel_name: str
    before: int | None
    after: int | None



"""Host a dumdum server.

To use TLS encryption, you must provide a certificate and private key.
This can be specified as either:
1. A single file containing both the private key and certificate:
     --cert hello.pem
2. A pair of certificate and private key files, separated with a colon:
     --cert hello.crt:hello.key

"""

import argparse
import asyncio
import ssl

from dumdum.protocol import Channel
from dumdum.logging import configure_logging

from .manager import host_server
from .state import ServerState


def main():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase logging verbosity",
    )
    parser.add_argument(
        "-c",
        "--channels",
        action="extend",
        nargs="+",
        help="A list of channels",
        type=parse_channel,
    )
    parser.add_argument(
        "--host",
        default=None,
        help="The address to host on",
    )
    parser.add_argument(
        "--port",
        default=6667,
        help="The port number to host on",
    )
    parser.add_argument(
        "--cert",
        default="",
        help="The SSL certificate and private key to use",
        type=parse_cert,
    )

    args = parser.parse_args()
    verbose: int = args.verbose
    channels: list[Channel] = args.channels or get_default_channels()
    host: str | None = args.host
    port: int = args.port
    ssl_context: ssl.SSLContext | None = args.cert

    configure_logging("server", verbose)

    state = ServerState()
    for channel in channels:
        state.add_channel(channel)

    try:
        asyncio.run(host_server(state, host, port, ssl=ssl_context))
    except KeyboardInterrupt:
        pass


def parse_channel(s: str) -> Channel:
    return Channel(name=s)


def get_default_channels() -> list[Channel]:
    return [Channel("general")]


def parse_cert(s: str) -> ssl.SSLContext | None:
    if s == "":
        return

    cafile, _, keyfile = s.partition(":")
    keyfile = keyfile or None

    context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
    context.load_cert_chain(cafile, keyfile)
    return context


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from typing import Self

from . import varchar
from .constants import MAX_CHANNEL_NAME_LENGTH
from .reader import Reader


@dataclass
class Channel:
    name: str = field(hash=True)

    def __bytes__(self) -> bytes:
        return varchar.dumps(self.name, max_length=MAX_CHANNEL_NAME_LENGTH)

    @classmethod
    def from_reader(cls, reader: Reader) -> Self:
        name = reader.read_varchar(max_length=MAX_CHANNEL_NAME_LENGTH)
        return cls(name=name)



import datetime
import json
import logging
import logging.config
from typing import Any, Literal

from .appdirs import APP_DIRS

BASE_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "root": {
        "level": "DEBUG",
        "handlers": ["stdout", "file"],
    },
    "handlers": {
        "stdout": {
            "class": "logging.StreamHandler",
            "formatter": "text",
            "stream": "ext://sys.stdout",
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "json",
            "filename": "",  # Filled in by configure_logging()
            "maxBytes": 5_000_000,
            "backupCount": 3,
        },
    },
    "formatters": {
        "text": {
            "format": "%(message)s",  # Filled in by configure_logging()
            "datefmt": "%Y-%m-%dT%H:%M:%S%z",
        },
        "json": {
            "()": "dumdum.logging.JSONFormatter",
        },
    },
}

LOG_RECORD_ATTRIBUTES = {
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",  # cached by formatException()
    "filename",
    "funcName",
    "levelname",
    "levelno",
    "lineno",
    "message",
    "module",
    "msecs",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "stack_info",
    "thread",
    "threadName",
    "taskName",
}


def configure_logging(mode: Literal["client", "server"], verbose: int) -> None:
    config = BASE_CONFIG.copy()

    if verbose == 0:
        text_format = "%(levelname)s: %(message)s"
        stdout_level = logging.WARNING
        file_level = logging.INFO
    elif verbose == 1:
        text_format = "%(levelname)s: %(message)s"
        stdout_level = logging.INFO
        file_level = logging.INFO
    else:
        text_format = "%(levelname)s: %(message)-50s (%(name)s#L%(lineno)d)"
        stdout_level = logging.DEBUG
        file_level = logging.DEBUG

    config["formatters"]["text"]["format"] = text_format
    config["handlers"]["stdout"]["level"] = stdout_level
    config["handlers"]["file"]["level"] = file_level

    if mode == "server":
        file_filename = str(APP_DIRS.user_log_path / "dumdum-server.jsonl")
    else:
        file_filename = str(APP_DIRS.user_log_path / "dumdum.jsonl")

    config["handlers"]["file"]["filename"] = file_filename

    logging.config.dictConfig(config)


class JSONFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        data = self._prepare_record(record)
        return json.dumps(data, default=str)

    def _prepare_record(self, record: logging.LogRecord) -> dict[str, Any]:
        data = {}

        for k, v in vars(record).items():
            if k not in LOG_RECORD_ATTRIBUTES:
                data[k] = v

        created = datetime.datetime.fromtimestamp(
            record.created,
            tz=datetime.timezone.utc,
        )

        data["created"] = created.isoformat()
        data["message"] = record.getMessage()

        if record.exc_info is not None:
            data["exc_info"] = self.formatException(record.exc_info)

        if record.stack_info is not None:
            data["stack_info"] = self.formatStack(record.stack_info)

        return data



from dataclasses import dataclass

from dumdum.protocol import varchar
from dumdum.protocol.constants import (
    MAX_CHANNEL_NAME_LENGTH,
    MAX_MESSAGE_LENGTH,
    MAX_NICK_LENGTH,
)
from dumdum.protocol.enums import ClientMessageType


@dataclass
class ClientMessageHello:
    version: int

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ClientMessageType.HELLO.value,
                self.version,
            ]
        )


@dataclass
class ClientMessageAuthenticate:
    nick: str

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ClientMessageType.AUTHENTICATE.value,
                *varchar.dumps(self.nick, max_length=MAX_NICK_LENGTH),
            ]
        )


@dataclass
class ClientMessagePost:
    channel_name: str
    content: str

    def __bytes__(self) -> bytes:
        return bytes(
            [
                ClientMessageType.SEND_MESSAGE.value,
                *varchar.dumps(self.channel_name, max_length=MAX_CHANNEL_NAME_LENGTH),
                *varchar.dumps(self.content, max_length=MAX_MESSAGE_LENGTH),
            ]
        )


@dataclass
class ClientMessageListChannels:
    def __bytes__(self) -> bytes:
        return bytes([ClientMessageType.LIST_CHANNELS.value])


@dataclass
class ClientMessageListMessages:
    channel_name: str
    before: int | None
    after: int | None

    def __bytes__(self) -> bytes:
        before = self.before or 0
        after = self.after or 0
        return bytes(
            [
                ClientMessageType.LIST_MESSAGES.value,
                *varchar.dumps(self.channel_name, max_length=MAX_CHANNEL_NAME_LENGTH),
                *before.to_bytes(8, byteorder="big"),
                *after.to_bytes(8, byteorder="big"),
            ]
        )



from dataclasses import dataclass
from typing import Sequence

from dumdum.protocol.channel import Channel
from dumdum.protocol.message import Message


@dataclass
class ClientEvent:
    """An event received by the client."""


@dataclass
class ClientEventHello(ClientEvent):
    """The server responded to our hello."""

    using_ssl: bool


@dataclass
class ClientEventIncompatibleVersion(ClientEvent):
    """The server does not support our protocol version."""

    server_version: int
    client_version: int


@dataclass
class ClientEventAuthentication(ClientEvent):
    """The server responded to our authentication request."""

    success: bool


@dataclass
class ClientEventMessageReceived(ClientEvent):
    """The server broadcasted a message to the client."""

    message: Message


@dataclass
class ClientEventChannelsListed(ClientEvent):
    """The server responded to our request for a channel list."""

    channels: Sequence[Channel]


@dataclass
class ClientEventMessagesListed(ClientEvent):
    """The server responded to our request for a message list."""

    messages: Sequence[Message]



from pathlib import Path
from tkinter import StringVar, filedialog
from tkinter.ttk import Button, Entry, Frame, Label
from typing import Any

ALL_FILETYPE = ("All files", ("*",))
SSL_CERTIFICATE_FILETYPE = (
    "SSL Certificate",
    ("*.pem", "*.crt", "*.ca-bundle", "*.cer"),
)


class FileEntry(Frame):
    def __init__(
        self,
        parent: Frame,
        *,
        browse_kwargs: dict[str, Any] | None = None,
        text: str = "",
    ) -> None:
        super().__init__(parent)

        if browse_kwargs is None:
            browse_kwargs = {}
        self.browse_kwargs = browse_kwargs

        self.grid_columnconfigure(1, weight=1)

        self._validatecommand = (self.register(self._validate), "%P")

        self.label = Label(self, text=text)
        self.label.grid(row=0, column=0)

        self.var = StringVar(self)
        self.entry = Entry(
            self,
            textvariable=self.var,
            validate="focusout",
            validatecommand=self._validatecommand,
        )
        self.entry.grid(row=0, column=1, sticky="ew")

        self.browse = Button(self, text="Browse", command=self.do_browse)
        self.browse.grid(row=0, column=2)

    def enable(self, enabled: bool) -> None:
        state = ["!disabled" if enabled else "disabled"]
        self.entry.state(state)
        self.browse.state(state)

    def get(self) -> str:
        return self.var.get().strip()

    def get_path(self) -> Path | None:
        filename = self.get()
        if filename != "":
            return Path(filename)

    def do_browse(self) -> None:
        filename = filedialog.askopenfilename(**self.browse_kwargs)
        if filename != "":
            self.var.set(filename)

    def _validate(self, val: str) -> bool:
        return Path(val).is_file()



from __future__ import annotations

import collections
import concurrent.futures
import time
from tkinter import Event, StringVar
from tkinter.ttk import Button, Entry, Frame, Label, Treeview
from typing import ContextManager

from dumdum.protocol import (
    Channel,
    ClientEvent,
    ClientEventChannelsListed,
    ClientEventMessageReceived,
    ClientEventMessagesListed,
    Message,
)

from .app import TkApp
from .scrollable_frame import ScrollableFrame
from .store import ClientStore


class ChatFrame(Frame):
    _connection_attempt: concurrent.futures.Future[None] | None

    def __init__(self, app: TkApp):
        super().__init__(padding=10)

        self.app = app

        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=4)
        self.grid_rowconfigure(0, weight=1)

        self.channels: list[Channel] = []
        self.message_cache = MessageCache()

        self.channel_list = ChannelList(self)
        self.channel_list.grid(row=0, column=0, rowspan=2, padx=(0, 10), sticky="nesw")
        self.messages = MessageList(self)
        self.messages.grid(row=0, column=1, sticky="nesw")
        self.send_box = SendBox(self)
        self.send_box.grid(row=1, column=1, sticky="nesw")

    def handle_client_event(self, event: ClientEvent) -> None:
        if isinstance(event, ClientEventChannelsListed):
            self.channels.clear()
            self.channels.extend(event.channels)
            self.channel_list.refresh()

            for channel in event.channels:
                coro = self.app.client.list_messages(channel.name)
                self.app.submit(coro)

        elif isinstance(event, ClientEventMessageReceived):
            self.add_message(event.message)
        elif isinstance(event, ClientEventMessagesListed):
            for message in event.messages:
                self.add_message(message)

    def add_message(self, message: Message) -> None:
        self.message_cache.add_message(message)
        channel = self.get_channel(message.channel_name)
        if channel == self.channel_list.selected_channel:
            self.messages.add_message(message)

    def get_channel(self, name: str) -> Channel | None:
        for channel in self.channels:
            if channel.name == name:
                return channel


class ChannelList(Frame):
    def __init__(self, parent: ChatFrame) -> None:
        super().__init__(parent)

        self.parent = parent

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.tree = Treeview(self, selectmode="browse", show="tree")
        self.tree.grid(sticky="nesw")

        # TODO: maybe add button to refresh channels?

        self.tree.bind("<<TreeviewSelect>>", self._on_tree_select)

    @property
    def store(self) -> ContextManager[ClientStore]:
        return self.parent.app.store_factory()

    @property
    def addr(self) -> str:
        return self.parent.app.client.addr

    @property
    def selected_channel(self) -> Channel | None:
        selection = self.tree.selection()
        if len(selection) < 1:
            return None

        return self.parent.get_channel(selection[0])

    def refresh(self) -> None:
        selection = self.tree.selection()
        self.tree.delete(*self.tree.get_children())

        for channel in self.parent.channels:
            self.tree.insert("", "end", channel.name, text=channel.name)

        children = self.tree.get_children()
        if len(selection) == 0 and len(children) > 0:
            with self.store as store:
                last = store.get_last_selected_channel(self.addr)
                if last is None or last not in children:
                    last = children[0]
                self.tree.selection_set(last)

    def _on_tree_select(self, event: Event) -> None:
        selected_channel = self.selected_channel

        with self.store as store:
            name = selected_channel and selected_channel.name
            store.set_last_selected_channel(self.addr, name)

        self.parent.messages.set_channel(selected_channel)


class MessageList(Frame):
    UPDATE_RATE = 250
    UPDATE_IDLE_PERIOD = 0.75

    _last_configured: float | None

    def __init__(self, parent: ChatFrame):
        super().__init__(parent)

        self.parent = parent

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.messages: list[MessageView] = []

        self._scroll_frame = ScrollableFrame(
            self,
            xscroll=True,
            yscroll=True,
            autoscroll=True,
        )
        self._scroll_frame.grid(row=0, column=0, sticky="nesw")

        self._last_configured = None
        self.after(self.UPDATE_RATE, self._update_loop)

        self.bind("<Configure>", self._on_configure)

    def add_message(self, message: Message) -> None:
        widget = MessageView(self._scroll_frame.inner, self, message)
        widget.wrap_to_width(self.winfo_width())
        widget.grid(row=len(self.messages), column=0, sticky="ew")
        self.messages.append(widget)

    def clear_messages(self) -> list[MessageView]:
        messages = self.messages.copy()
        self.messages.clear()

        for message in messages:
            message.grid_forget()

        return messages

    def set_channel(self, channel: Channel | None) -> None:
        self.clear_messages()

        if channel is not None:
            for message in self.parent.message_cache.get_messages(channel):
                self.add_message(message)

    def _on_configure(self, event: Event) -> None:
        self._last_configured = time.perf_counter()

    def _update_loop(self) -> None:
        self.after(self.UPDATE_RATE, self._update_loop)

        if self._last_configured is None:
            return
        elif time.perf_counter() < self._last_configured + self.UPDATE_IDLE_PERIOD:
            return

        self._last_configured = None
        for message in self.messages:
            message.wrap_to_width(self.winfo_width())


class MessageView(Frame):
    def __init__(self, frame: Frame, parent: MessageList, message: Message) -> None:
        super().__init__(frame)

        self.parent = parent
        self.message = message

        self.nick = Label(self, text=f"{message.nick}:", font="bold 10")
        self.nick.grid(row=0, column=0, sticky="nw")
        self.content = Label(self, text=message.content, wraplength=1000)
        self.content.grid(row=0, column=1, sticky="ew")

    def wrap_to_width(self, width: int) -> None:
        minimum = 800
        maximum = width - self.nick.winfo_reqwidth() - 20
        self.content.configure(wraplength=max(minimum, maximum))


class MessageCache:
    def __init__(self) -> None:
        self.channel_messages: dict[str, list[Message]] = collections.defaultdict(list)

    def add_message(self, message: Message) -> None:
        self.channel_messages[message.channel_name].append(message)

    def get_messages(self, channel: Channel) -> list[Message]:
        messages = self.channel_messages.get(channel.name)
        if messages is None:
            return []
        return messages.copy()


class SendBox(Frame):
    def __init__(self, parent: ChatFrame) -> None:
        super().__init__(parent)

        self.parent = parent

        self.grid_columnconfigure(0, weight=1)

        self.content_entry_var = StringVar(self)
        self.content_entry = Entry(self, textvariable=self.content_entry_var)
        self.content_entry.grid(row=0, column=0, sticky="ew")
        self.send_button = Button(self, text="Send", command=self.do_send)
        self.send_button.grid(row=0, column=1)

        self.content_entry.bind("<Return>", lambda event: self.do_send())

    def do_send(self) -> None:
        content = self.content_entry_var.get()
        if len(content) < 1:
            return

        channel = self.parent.channel_list.selected_channel
        if channel is None:
            return

        self.content_entry_var.set("")
        coro = self.parent.app.client.send_message(channel.name, content)
        self.parent.app.submit(coro)



"""Start the client interface for connecting to dumdum servers."""

from __future__ import annotations

import argparse
import contextlib
import sys
from typing import Iterator

from dumdum.logging import configure_logging

from .app import TkApp
from .connect_frame import ConnectFrame
from .event_thread import EventThread
from .store import ClientStore


def main():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase logging verbosity",
    )
    parser.set_defaults(mode="gui")

    commands = parser.add_subparsers()

    appdirs = commands.add_parser(
        "appdirs",
        description="Show data directories used by dumdum.",
    )
    appdirs.set_defaults(mode="appdirs")

    args = parser.parse_args()
    verbose: int = args.verbose
    mode: str = args.mode

    configure_logging("client", verbose)
    enable_windows_dpi_awareness()

    if mode == "gui":
        with contextlib.suppress(KeyboardInterrupt):
            run_gui()
    elif mode == "appdirs":
        show_appdirs()
    else:
        raise RuntimeError(f"Unknown mode {mode!r}")


def run_gui() -> None:
    with EventThread() as event_thread:
        app = TkApp(
            event_thread=event_thread,
            store_factory=store_factory,
        )
        app.switch_frame(ConnectFrame(app))

        try:
            app.mainloop()
        except BaseException:
            app.destroy()
            app.mainloop()
            raise


def enable_windows_dpi_awareness():
    if sys.platform == "win32":
        from ctypes import windll

        windll.shcore.SetProcessDpiAwareness(2)


def show_appdirs() -> None:
    from dumdum.appdirs import APP_DIRS

    print("user_data_path =", APP_DIRS.user_data_path)
    print("user_log_path =", APP_DIRS.user_log_path)


@contextlib.contextmanager
def store_factory() -> Iterator[ClientStore]:
    with ClientStore.from_appdirs() as store, store.transaction():
        yield store


if __name__ == "__main__":
    main()



"""
64 63                                           19     12           0
 0  00011000111000111000000101110101101101000100 1110101 000000000111

64-63: Always 0
63-19: Unix timestamp in milliseconds
19-12: Process ID
12-00: Incrementing per-process ID
"""

import datetime
import os
import time

_incrementing_id = 0


def create_snowflake(
    t: float | int | datetime.datetime | None = None,
    pid: int | None = None,
    increment: int | None = None,
) -> int:
    if t is None:
        # WARNING: this may roll back or forwards depending on system time
        t = time.time_ns() // 1000000
    elif isinstance(t, datetime.datetime):
        t = int(t.timestamp() * 1000)
    elif isinstance(t, float):
        t = int(t * 1000)

    if t.bit_length() > 44:
        raise OverflowError(f"Timestamp {t} out of bounds (are we in 2527?)")

    if pid is None:
        pid = os.getpid()

    if increment is None:
        global _incrementing_id
        increment = _incrementing_id
        _incrementing_id += 1

    pid = pid % 128
    increment = increment % 4096

    return (t << 19) + (pid << 12) + increment



import pytest

from dumdum.protocol import InvalidLengthError, byte_reader, varchar


def test_loads_unicode():
    message = b"\x17foobar2000 hamspam \xf0\x9f\x91\x80"
    assert varchar.loads(message, max_length=23) == "foobar2000 hamspam 👀"


def test_loads_extra_data():
    message = b"\x05Hello world!\n"
    assert varchar.loads(message, max_length=5) == "Hello"
    assert varchar.loads(message, max_length=13) == "Hello"


def test_loads_exceeds_max_length():
    message = b"\x05Hello world!\n"
    with pytest.raises(InvalidLengthError):
        varchar.loads(message, max_length=3)


def test_loads_insufficient_data():
    message = b"\xffThis message is too short!"
    with pytest.raises(IndexError):
        varchar.loads(message, max_length=255)


def test_loads_insufficient_length():
    message = b"\x01"
    with pytest.raises(IndexError):
        varchar.loads(message, max_length=256)


def test_loads_empty_message():
    message = b""
    assert varchar.loads(message, max_length=0) == ""
    with pytest.raises(IndexError):
        varchar.loads(message, max_length=1)


def test_loads_multi_byte_length():
    message = b"\x01\xff" + b"x" * (256 + 255)
    assert varchar.loads(message, max_length=511) == "x" * 511
    assert varchar.loads(message, max_length=65535) == "x" * 511

    with pytest.raises(InvalidLengthError):
        expected_length = int.from_bytes(message[:3], byteorder="big")
        assert expected_length > 65536
        assert varchar.loads(message, max_length=65536)


def test_load():
    with byte_reader(b"\x17foobar2000 hamspam \xf0\x9f\x91\x80") as reader:
        assert varchar.load(reader, max_length=23) == "foobar2000 hamspam 👀"


def test_dumps_unicode():
    message = "foobar2000 hamspam 👀"
    expected = b"\x17foobar2000 hamspam \xf0\x9f\x91\x80"
    assert varchar.dumps(message, max_length=len(message.encode())) == expected

    with pytest.raises(InvalidLengthError):
        varchar.dumps(message, max_length=len(message))


def test_dumps_empty_message():
    for i in range(5):
        length = 2 ** (8 * i)
        assert varchar.dumps("", max_length=length) == b"\x00" * (i + 1)



from tkinter import Canvas, Event, Widget
from tkinter.ttk import Frame, Scrollbar, Style
from weakref import WeakSet


class ScrollableFrame(Frame):
    __last_scrollregion: tuple[int, int, int, int] | None

    def __init__(
        self,
        *args,
        autoscroll: bool = False,
        xscroll: bool = False,
        yscroll: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.autoscroll = autoscroll
        self.xscroll = xscroll
        self.yscroll = yscroll

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.__canvas = Canvas(self, highlightthickness=0)
        self.__canvas.grid(row=0, column=0, sticky="nesw")

        self.__xscrollbar = Scrollbar(self, orient="horizontal")
        self.__yscrollbar = Scrollbar(self, orient="vertical")

        # Use rowspan=2 or columnspan=2 appropriately if filling the
        # bottom-right corner of the frame is desired.
        self.__xscrollbar.grid(row=1, column=0, sticky="ew")
        self.__yscrollbar.grid(row=0, column=1, sticky="ns")

        self.__xscrollbar["command"] = self.__canvas.xview
        self.__yscrollbar["command"] = self.__canvas.yview
        self.__canvas["xscrollcommand"] = self.__wrap_scrollbar_set(self.__xscrollbar)
        self.__canvas["yscrollcommand"] = self.__wrap_scrollbar_set(self.__yscrollbar)

        self.inner = Frame(self.__canvas)
        self.__inner_id = self.__canvas.create_window(
            (0, 0), window=self.inner, anchor="nw"
        )

        self.__canvas.bind("<Configure>", lambda event: self.__update())
        self.inner.bind("<Configure>", self.__on_inner_configure)

        self.__last_scrollregion = None
        self.__scrolled_widgets = WeakSet()
        self.__style = Style(self)
        self.__update_rate = 125
        self.__update_loop()

    def __on_inner_configure(self, event: Event):
        background = self.__style.lookup(self.inner.winfo_class(), "background")
        self.__canvas.configure(background=background)
        self.__update()

    def __update_loop(self):
        # Without this, any changes to the inner frame won't affect
        # the scroll bar/region until the window is resized.
        self.__update()
        self.after(self.__update_rate, self.__update_loop)

    def __update(self):
        scroll_edges = self.__get_scroll_edges()

        # self._canvas.bbox("all") doesn't update until window resize
        # so we're relying on the inner frame's requested height instead.
        new_width = self.__canvas.winfo_width()
        new_height = self.__canvas.winfo_height()
        if self.xscroll:
            new_width = max(new_width, self.inner.winfo_reqwidth())
        if self.yscroll:
            new_height = max(new_height, self.inner.winfo_reqheight())

        bbox = (0, 0, new_width, new_height)
        self.__canvas.configure(scrollregion=bbox)
        self.__canvas.itemconfigure(self.__inner_id, width=new_width, height=new_height)

        self.__update_scrollbar_visibility(self.__xscrollbar)
        self.__update_scrollbar_visibility(self.__yscrollbar)
        self.__propagate_scroll_binds(self.inner)
        self.__update_scroll_edges(bbox, *scroll_edges)

    def __get_scroll_edges(self) -> tuple[bool, bool]:
        xview = self.__canvas.xview()
        yview = self.__canvas.yview()
        scrolled_to_right = xview[1] == 1 and xview[0] != 0
        scrolled_to_bottom = yview[1] == 1 and yview[0] != 0
        return scrolled_to_right, scrolled_to_bottom

    def __propagate_scroll_binds(self, parent: Widget):
        if parent not in self.__scrolled_widgets:
            parent.bind("<MouseWheel>", self.__on_mouse_yscroll)
            parent.bind("<Shift-MouseWheel>", self.__on_mouse_xscroll)
            self.__scrolled_widgets.add(parent)

        for child in parent.winfo_children():
            self.__propagate_scroll_binds(child)

    def __update_scroll_edges(
        self,
        bbox: tuple[int, int, int, int],
        scrolled_to_right: bool,
        scrolled_to_bottom: bool,
    ) -> None:
        self.__last_scrollregion, last_bbox = bbox, self.__last_scrollregion
        if not self.autoscroll:
            return
        elif bbox == last_bbox:
            return

        if scrolled_to_right:
            self.__canvas.xview_moveto(1)
        if scrolled_to_bottom:
            self.__canvas.yview_moveto(1)

    def __on_mouse_xscroll(self, event: Event):
        delta = int(-event.delta / 100)
        self.__canvas.xview_scroll(delta, "units")

    def __on_mouse_yscroll(self, event: Event):
        delta = int(-event.delta / 100)
        self.__canvas.yview_scroll(delta, "units")

    def __update_scrollbar_visibility(self, scrollbar: Scrollbar):
        if scrollbar.get() == (0, 1):
            scrollbar.grid_remove()
        else:
            scrollbar.grid()

    def __wrap_scrollbar_set(self, scrollbar: Scrollbar):
        def wrapper(*args, **kwargs):
            scrollbar.set(*args, **kwargs)
            self.__update_scrollbar_visibility(scrollbar)

        return wrapper



import asyncio
import concurrent.futures
import threading
from typing import Self


class EventThread(threading.Thread):
    """Handles starting and stopping an asyncio event loop in a separate thread."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.loop_fut = concurrent.futures.Future()
        self.stop_fut = concurrent.futures.Future()
        self.finished_fut = concurrent.futures.Future()

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, tb) -> None:
        self.stop()
        self.join()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self.loop_fut.result()

    def run(self) -> None:
        try:
            asyncio.run(self._run_forever())
        finally:
            self.finished_fut.set_result(None)

    def stop(self) -> None:
        if not self.stop_fut.done():
            self.stop_fut.set_result(None)

    async def _run_forever(self) -> None:
        self.loop_fut.set_result(asyncio.get_running_loop())
        await asyncio.wrap_future(self.stop_fut)



from .client import (
    Client,
    ClientEvent,
    ClientEventAuthentication,
    ClientEventChannelsListed,
    ClientEventHello,
    ClientEventIncompatibleVersion,
    ClientEventMessageReceived,
    ClientEventMessagesListed,
    ClientMessageAuthenticate,
    ClientMessageHello,
    ClientMessageListChannels,
    ClientMessageListMessages,
    ClientMessagePost,
    ClientState,
)
from .server import (
    Server,
    ServerEvent,
    ServerEventAuthentication,
    ServerEventHello,
    ServerEventIncompatibleVersion,
    ServerEventListChannels,
    ServerEventListMessages,
    ServerEventMessageReceived,
    ServerMessageAcknowledgeAuthentication,
    ServerMessageHello,
    ServerMessageListChannels,
    ServerMessageListMessages,
    ServerMessagePost,
    ServerMessageSendIncompatibleVersion,
    ServerState,
)
from .buffer import extend_limited_buffer
from .channel import Channel
from .constants import MAX_MESSAGE_LENGTH, MAX_NICK_LENGTH
from .enums import ClientMessageType, ServerMessageType
from .errors import (
    BufferOverflowError,
    InvalidLengthError,
    InvalidStateError,
    MalformedDataError,
    ProtocolError,
)
from .interfaces import Protocol
from .message import Message
from .reader import Reader, bytearray_reader, byte_reader
from .snowflake import create_snowflake



import bisect
import collections
from typing import Sequence, TypeAlias

from dumdum.protocol import Channel, Message

User: TypeAlias = str


class ServerState:
    def __init__(self) -> None:
        self._channels: dict[str, Channel] = {}
        self._messages: dict[str, list[Message]] = collections.defaultdict(list)
        self._users: dict[str, User] = {}

    @property
    def channels(self) -> tuple[Channel, ...]:
        return tuple(self._channels.values())

    def add_channel(self, channel: Channel) -> None:
        self._channels[channel.name] = channel

    def get_channel(self, name: str) -> Channel | None:
        return self._channels.get(name)

    def remove_channel(self, name: str) -> Channel | None:
        return self._channels.pop(name, None)

    def get_messages(
        self,
        channel_name: str,
        *,
        before: int | None = None,
        after: int | None = None,
    ) -> Sequence[Message]:
        messages = self._messages[channel_name]

        if before is not None:
            i = bisect.bisect_left(messages, before, key=lambda m: m.id)
            messages = messages[i:]

        if after is not None:
            i = bisect.bisect_right(messages, after, key=lambda m: m.id)
            messages = messages[:i]

        return messages[-100:]

    def add_message(self, message: Message) -> None:
        messages = self._messages[message.channel_name]
        bisect.insort(messages, message, key=lambda m: m.id)

    def get_message(self, channel_name: str, id: int) -> Message | None:
        messages = self._messages[channel_name]
        if len(messages) < 1:
            return None

        i = self._index_message(messages, id)
        message = messages[i]
        if message.id == id:
            return message

    def remove_message(self, channel_name: str, id: int) -> Message | None:
        messages = self._messages[channel_name]
        if len(messages) < 1:
            return None

        i = self._index_message(messages, id)
        message = messages[i]
        if message.id == id:
            del messages[i]
            return message

    def _index_message(self, messages: Sequence[Message], id: int) -> int:
        return bisect.bisect_left(messages, id, key=lambda m: m.id)

    @property
    def users(self) -> tuple[User, ...]:
        return tuple(self._users.values())

    def add_user(self, user: User) -> None:
        self._users[user] = user

    def get_user(self, nick: str) -> User | None:
        return self._users.get(nick)

    def remove_user(self, nick: str) -> User | None:
        return self._users.pop(nick, None)



import logging
import re
from typing import Literal

import importlib_resources

from .connection import SQLiteConnection

log = logging.getLogger(__name__)

MigrationPath = Literal["client"]


def run_migrations(conn: SQLiteConnection, path: MigrationPath):
    version: int = conn.fetchval("PRAGMA user_version")
    script_version = version

    for script_version, script in _get_migrations(path, version).items():
        conn.executescript(script)

    if script_version > version:
        conn.execute(f"PRAGMA user_version = {script_version}")
        log.debug("Migrated database v%d to v%d", version, script_version)


def _get_migrations(path: MigrationPath, version: int) -> dict[int, str]:
    migrations: dict[int, str] = {}

    resource = importlib_resources.files(__package__) / "scripts" / path
    with importlib_resources.as_file(resource) as root:
        for file in root.glob("*.sql"):
            m = re.match(r"\d+", file.stem)
            if m is None:
                continue

            file_version = int(m[0])
            if file_version in migrations:
                raise ValueError(
                    f"Duplicate migration found for version {file_version}"
                )

            if file_version <= version:
                continue

            migrations[file_version] = file.read_text("utf-8")

    return dict(sorted(migrations.items()))



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from dumdum.protocol import Server, ServerEvent

if TYPE_CHECKING:
    from .manager import Manager


class Connection:
    nick: str | None

    def __init__(
        self,
        manager: Manager,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
        server: Server,
    ):
        self.manager = manager
        self.reader = reader
        self.writer = writer
        self.server = server

        self.nick = None

    async def communicate(self) -> None:
        while True:
            data = await self.reader.read(1024)
            if len(data) == 0:
                break

            events, outgoing = self.server.receive_bytes(data)
            self.writer.write(outgoing)
            await self._handle_events(events)
            await self._drain()  # exert backpressure

    async def _handle_events(self, events: list[ServerEvent]) -> None:
        await self.manager._handle_events(self, events)

    async def _drain(self) -> None:
        timeout = self.manager.drain_timeout
        await asyncio.wait_for(self.writer.drain(), timeout=timeout)



from typing import Sequence, Type, TypeVar

import pytest

from dumdum.protocol import (
    BufferOverflowError,
    Channel,
    Client,
    ClientEventAuthentication,
    ClientEventChannelsListed,
    ClientEventHello,
    ClientEventIncompatibleVersion,
    ClientEventMessagesListed,
    ClientState,
    InvalidStateError,
    MalformedDataError,
    Message,
    Protocol,
    Server,
    ServerEventAuthentication,
    ServerEventHello,
    ServerEventIncompatibleVersion,
    ServerEventListChannels,
    ServerEventListMessages,
    ServerEventMessageReceived,
    ServerState,
)

T = TypeVar("T")


def communicate(
    sender: Protocol,
    data: bytes,
    receiver: Protocol,
) -> tuple[list[object], list[object]]:
    def send_loop():
        i, outgoing = 0, data
        while len(outgoing) > 0:
            receiver, receiver_events = order[i % len(order)]
            print(f"{type(receiver).__name__} << {outgoing}")
            events, outgoing = receiver.receive_bytes(outgoing)
            receiver_events.extend(events)
            i += 1

    sender_events: list[object] = []
    receiver_events: list[object] = []
    order = [(receiver, receiver_events), (sender, sender_events)]
    send_loop()
    return sender_events, receiver_events


def assert_event_order(events: Sequence[object], *expected_events: Type[object]):
    # fmt: off
    assert len(events) == len(expected_events), \
        "Expected exactly {} event(s), got {} instead".format(
        len(expected_events),
        len(events),
    )

    for i, (event, type_) in enumerate(zip(events, expected_events)):
        assert isinstance(event, type_), \
            "Expected {} at index {}, got {} instead".format(
            type_.__name__,
            i,
            type(event).__name__,
        )
    # fmt: on


def test_authenticate():
    nick = "thegamecracks"
    client = Client(nick=nick)
    server = Server()

    client_events, server_events = communicate(client, client.hello(), server)
    assert client_events == []
    assert server_events == [ServerEventHello()]

    data = server.hello(using_ssl=False)
    server_events, client_events = communicate(server, data, client)
    assert client_events == [ClientEventHello(using_ssl=False)]
    assert server_events == []

    client_events, server_events = communicate(client, client.authenticate(), server)
    assert client_events == []
    assert server_events == [ServerEventAuthentication(nick=nick)]

    data = server.authenticate(success=True)
    server_events, client_events = communicate(server, data, client)
    assert client_events == [ClientEventAuthentication(success=True)]
    assert server_events == []


def test_authenticate_incompatible_version():
    nick = "thegamecracks"
    client = Client(nick=nick)
    server = Server()

    client.PROTOCOL_VERSION = server.PROTOCOL_VERSION + 1  # type: ignore
    if client.PROTOCOL_VERSION > 255:
        client.PROTOCOL_VERSION = server.PROTOCOL_VERSION - 1  # type: ignore

    assert client.PROTOCOL_VERSION != server.PROTOCOL_VERSION

    client_events, server_events = communicate(client, client.hello(), server)
    assert client_events == [
        ClientEventIncompatibleVersion(
            client_version=client.PROTOCOL_VERSION,
            server_version=server.PROTOCOL_VERSION,
        )
    ]
    assert server_events == [
        ServerEventIncompatibleVersion(
            version=client.PROTOCOL_VERSION,
        )
    ]


def test_send_message():
    nick = "thegamecracks"
    content = "Hello world!"
    channel_name = "general"

    client = Client(nick=nick)
    server = Server()

    communicate(client, client.hello(), server)
    communicate(server, server.hello(using_ssl=False), client)
    communicate(client, client.authenticate(), server)
    communicate(server, server.authenticate(success=True), client)

    client_events, server_events = communicate(
        client,
        client.send_message(channel_name, content),
        server,
    )
    assert client_events == []
    assert server_events == [ServerEventMessageReceived(channel_name, content)]


def test_list_channels():
    channels = [
        Channel("announcements"),
        Channel("general"),
        Channel("memes"),
        Channel("coding-lounge"),
        Channel("fishing-trips"),
    ]

    client = Client(nick="thegamecracks")
    server = Server()

    communicate(client, client.hello(), server)
    communicate(server, server.hello(using_ssl=False), client)
    communicate(client, client.authenticate(), server)
    communicate(server, server.authenticate(success=True), client)

    client_events, server_events = communicate(client, client.list_channels(), server)
    assert client_events == []
    assert server_events == [ServerEventListChannels()]

    data = server.list_channels(channels)
    server_events, client_events = communicate(server, data, client)
    assert client_events == [ClientEventChannelsListed(channels)]
    assert server_events == []


def test_unauthenticated_send_message():
    nick = "thegamecracks"
    content = "Hello world!"
    channel = Channel("general")

    client = Client(nick=nick)
    server = Server()

    with pytest.raises(InvalidStateError):
        client.send_message(channel.name, content)

    client._state = ClientState.READY
    data = client.send_message(channel.name, content)

    with pytest.raises(InvalidStateError):
        communicate(client, data, server)


def test_unauthenticated_server_send_message():
    nick = "thegamecracks"
    content = "Hello world!"
    message = Message(0, "general", nick, content)

    client = Client(nick=nick)
    server = Server()

    with pytest.raises(InvalidStateError):
        server.send_message(message)

    server._state = ServerState.READY
    data = server.send_message(message)

    with pytest.raises(InvalidStateError):
        communicate(server, data, client)


def test_list_messages():
    nick = "thegamecracks"
    channel = Channel("general")

    client = Client(nick=nick)
    server = Server()

    communicate(client, client.hello(), server)
    communicate(server, server.hello(using_ssl=False), client)
    communicate(client, client.authenticate(), server)
    communicate(server, server.authenticate(success=True), client)

    before = 1
    after = 2
    data = client.list_messages(channel.name, before=before, after=after)
    client_events, server_events = communicate(client, data, server)
    assert client_events == []
    assert server_events == [ServerEventListMessages(channel.name, before, after)]

    messages = [Message(i, channel.name, nick, "Hello world!") for i in range(100)]
    data = server.list_messages(messages)
    server_events, client_events = communicate(server, data, client)
    assert server_events == []
    assert client_events == [ClientEventMessagesListed(messages)]


def test_invalid_message_type():
    client = Client("thegamecracks")
    server = Server()

    with pytest.raises(MalformedDataError):
        client.receive_bytes(b"\xff")

    with pytest.raises(MalformedDataError):
        server.receive_bytes(b"\xff")


def test_unicode_decode_error():
    client = Client("thegamecracks")
    server = Server()

    communicate(client, client.hello(), server)
    communicate(server, server.hello(using_ssl=False), client)
    communicate(client, client.authenticate(), server)
    communicate(server, server.authenticate(success=True), client)

    with pytest.raises(MalformedDataError):
        # LIST_CHANNELS, Channel name \N{EYES} but missing last 3 bytes
        data = b"\x04\x00\x02\x01\xf0"
        client.receive_bytes(data)

    with pytest.raises(MalformedDataError):
        # SEND_MESSAGE, Channel name \N{EYES} but missing last 3 bytes
        data = b"\x03\x01\xf0"
        server.receive_bytes(data)


def test_buffer_overflow_prevention():
    data = b"Hello world!\n"
    size = len(data) - 1

    client = Client("thegamecracks", buffer_size=size)
    server = Server(buffer_size=size)

    with pytest.raises(BufferOverflowError):
        client.receive_bytes(data)

    with pytest.raises(BufferOverflowError):
        server.receive_bytes(data)
